/*
 * File:   motorboard.c
 * Author: toffi
 *
 * Created on April 16, 2018, 11:52 PM
 */


#include <xc.h>
#include "pic_config.h"
#include "sw_config.h"
#include "led.h"

#define BOARD_ADDRESS 2 // for the UART playground

void main(void)
{
    // ========================================================================
    // Step 0: Oscillator Switch to get 140 MHz / 70 MIPS
    // No error handling here, if the oscillator fails we're fucked anyway
    // We have an 8MHz resonator, so we transform 8 -> 4 -> 280 -> 140 MHz
    PLLFBD = 68; // set PLL divider to 70 (from 4 to 280 MHz)
    CLKDIV = 0; // set N1 and N2 to 2 (8 to 4 and 280 to 140 MHz)
    __builtin_write_OSCCONH(0x03); // prepare clock switch to XTPLL
    __builtin_write_OSCCONL(OSCCON | 0x01); // initiate clock switch
    while(OSCCONbits.COSC != 0b011); // wait for XT to stabilize
    while(OSCCONbits.LOCK != 1); // wait for PLL lock
    // If we make it until here, the PIC is running at 70 MIPS.
    
    // ========================================================================
    // Step 1: I/O configuration
    // Normal Port Pin Configuration
    ANSELA = 0x0001; // AN0 is used for motor current measurement, all others off
    TRISA = 0xFFFF;
    TRISAbits.TRISA1 = 0; // RA1 is an output (LED red)
    
    ANSELB = 0; // no analog functions of port B used
    LATB = 0x0080; // ~CS of the encoder is high
    TRISB = 0xFFFF;
    TRISBbits.TRISB0 = 0; // RB0 (LED green)
    TRISBbits.TRISB1 = 0; // RB1 (LED blue)
    TRISBbits.TRISB7 = 0; // RB7 (Encoder ~CS
    TRISBbits.TRISB13 = 0; // RB13 (Motor PWM)
    TRISBbits.TRISB14 = 0; // RB14 (Motor direction)
    
    // Peripheral Pin Select Configuration
    // unlock PPS registers
    __builtin_write_OSCCONL(OSCCON & ~(1<<6));
    // CAN1 Module
    
    // UART1 module for networking / communications assignment
    RPINR18bits.U1RXR = 44; // Rx on RPI44 / RB12 -> J4
    RPOR4bits.RP43R = 1; // Tx on RP42 / RB11 -> J5
    
    // lock PPS registers
    __builtin_write_OSCCONL(OSCCON | (1<<6));
    
    // ========================================================================
    // Step 2: UART configuration
    U1BRG = 455; // 9,6k
    U1MODEbits.PDSEL = 0; // 8 data bits, no parity
    U1MODEbits.STSEL = 0; // 1 stop bit
    IPC3bits.U1TXIP = 5; // set priority level to 5 (arbitrary value)
    IPC2bits.U1RXIP = 5;
    U1MODEbits.UARTEN = 1; // enable the UART
    //U1STAbits.UTXISEL1 = 1; // interrupt whenever the transmit buffer is empty
    U1STAbits.URXISEL = 0; // interrupt whenever a character is received
    U1STAbits.UTXEN = 1; // enable transmissions
    IEC0bits.U1RXIE = 1; // enable Rx interrupt
    
    // ========================================================================
    // Step 3: System timer configuration
    // we use Timer2 in 16 bit mode
    T2CONbits.TCKPS0 = 1; // prescaler dividing by 8 -> counting @ 8750kHz
    PR2 = 8749; // counting from 0 to 8749 -> interrupts @ 1kHz
    IPC1bits.T2IP = 4; // lower priority than the UART
    IEC0bits.T2IE = 1; // enable timer 2 interrupt
    T2CONbits.TON = 1; // start timer
    
    // ========================================================================
    // Step 4: Interrupt configuration
    INTCON2bits.GIE = 1; // enable interrupts
    
    while(1);
}

// UART 1 Rx interrupt service routine
void __attribute__((__interrupt__, __auto_psv__)) _U1RXInterrupt(void)
{
    unsigned char receivedChar;
    static unsigned int foreign_message_counter = 0;
    static unsigned char ledState = 0;
    
    receivedChar = U1RXREG; // read from UART
    if(receivedChar >> 4 == BOARD_ADDRESS) // message is for us
    {
        if((receivedChar & 0x0F) == 1) // valid command in there
        {
            ledState++;
            if(ledState > 4) ledState = 0;
            setLED(WHITE, OFF);
            switch(ledState)
            {
                case 1:
                    setLED(RED, ON);
                    break;
                case 2:
                    setLED(GREEN, ON);
                    break;
                case 3:
                    setLED(BLUE, ON);
                    break;
                case 4:
                    setLED(YELLOW, ON);
                    break;
            }
        }
        foreign_message_counter = 0;
    }
    else // message is not for us, forward it
    {
        if(foreign_message_counter < 500) // only forward if the last message for us was less than 500 messages ago
        {
            U1TXREG = (unsigned int) receivedChar;
            foreign_message_counter++;
        }
    }
        
    IFS0bits.U1RXIF = 0; // reset interrupt flag
}

// Timer 2 interrupt service routine
// is called every 1ms
void __attribute__((__interrupt__, __auto_psv__)) _T2Interrupt(void)
{
    // Nothing here yet
    // end of timer 2 interrupt routine
    IFS0bits.T2IF = 0; // reset interrupt flag
}